@@ -16,7 +16,7 @@ use strict;
 use warnings;
 no warnings 'once';
 
-our $VERSION = '0.008000';
+our $VERSION = '0.009000';
 $VERSION = eval $VERSION;
 
 use Carp ();
@@ -46,41 +46,87 @@ our %NoTrace;
 $NoTrace{'Throwable::Error'}++;
 $NoTrace{'Moose::Error::Default'}++;
 
-our %OPTIONS;
+our %OPTIONS = (
+  objects   => !!1,
+  builtin   => undef,
+  dump      => !!0,
+  color     => !!0,
+  source    => 0,
+  evalsource => 0,
+  errors    => !!1,
+  warnings  => !!1,
+  better_names => !!1,
+);
+our %ENABLEOPTS = (
+  dump => 3,
+  source => 3,
+  evalsource => 3,
+);
+our %NUMOPTS = (
+  dump => 1,
+  source => 1,
+  evalsource => 1,
+);
+
+our @options = sort keys %OPTIONS;
+our ($opt_match) =
+  map qr/^-?(?:(no[_-]?)(?:$_)|(?:$_)(?:(\d+)|=(.*)|))$/,
+  join '|',
+  map {
+    my $o = $_;
+    $o =~ s/_/[-_]?/g;
+    '('.$o.')';
+  }
+  @options;
 
 sub _parse_options {
-  my @opts = map { /^-?(no[_-])?(.*)/; [ $_, $2, $1 ? 0 : 1 ] } @_;
-  if (!keys %OPTIONS) {
-    %OPTIONS = (
-      objects   => 1,
-      builtin   => undef,
-      dump      => 0,
-      color     => 0,
-      source    => 0,
-      errors    => 1,
-      warnings  => 1,
-      better_names => 1,
-    );
-    local $@;
-    eval {
-      _parse_options(
-        grep length, split /[\s,]+/, $ENV{DEVEL_CONFESS_OPTIONS}||''
-      );
-    } or warn "DEVEL_CONFESS_OPTIONS: $@";
-  }
-  for my $opt (@opts) {
-    if ($opt->[1] =~ /^dump(\d*)$/) {
-      $opt->[1] = 'dump';
-      $opt->[2] = length $1 ? ($1 || 'inf') : 3;
+  my %opts;
+  my @bad;
+  while (@_) {
+    my $arg = shift;
+    my @match = defined $arg ? $arg =~ $opt_match : ();
+    if (@match) {
+      my $no = shift @match;
+      my $equal = pop @match;
+      my $num = pop @match;
+      my ($opt) =
+        map $options[$_ % @options],
+        grep defined $match[$_],
+        0 .. $#match;
+      my $value
+        = defined $no       ? !!0
+        : defined $equal    ? $equal
+        : defined $num      ? $num
+        : @_ && (!defined $_[0] || $_[0] =~ /^\d+$/) ? shift
+        : defined $ENABLEOPTS{$opt} ? $ENABLEOPTS{$opt}
+        : !!1;
+
+      if ($NUMOPTS{$opt}) {
+        $value
+          = !defined $value ? 0
+          : !$value ? 0+'inf'
+          : 0+$value;
+      }
+      $opts{$opt} = $value;
+    }
+    else {
+      push @bad, $arg;
     }
   }
-  if (my @bad = grep { !exists $OPTIONS{$_->[1]} } @opts) {
+  if (@bad) {
     local $SIG{__DIE__};
-    Carp::croak("invalid options: " . join(', ', map { $_->[0] } @bad));
+    Carp::croak("invalid options: " . join(', ', map { defined $_ ? $_ : '[undef]' } @bad));
   }
-  $OPTIONS{$_->[1]} = $_->[2]
-    for @opts;
-  1;
+  \%opts;
+}
+
+if (my $env = $ENV{DEVEL_CONFESS_OPTIONS}) {
+  local $@;
+  eval {
+    my $options = _parse_options(grep length, split /[\s,]+/, $env);
+    @OPTIONS{keys %$options} = values %$options;
+    1;
+  } or warn "DEVEL_CONFESS_OPTIONS: $@";
 }
 
 our %OLD_SIG;
@@ -88,15 +134,17 @@ our %OLD_SIG;
 sub import {
   my $class = shift;
 
-  _parse_options(@_);
+  my $options = _parse_options(@_);
+  @OPTIONS{keys %$options} = values %$options;
 
   if (defined $OPTIONS{builtin}) {
     require Devel::Confess::Builtin;
     my $do = $OPTIONS{builtin} ? 'import' : 'unimport';
     Devel::Confess::Builtin->$do;
   }
-  if ($OPTIONS{source}) {
+  if ($OPTIONS{source} || $OPTIONS{evalsource}) {
     require Devel::Confess::Source;
+    Devel::Confess::Source->import;
   }
   if ($OPTIONS{color} && $^O eq 'MSWin32') {
     if (eval { require Win32::Console::ANSI }) {
@@ -156,11 +204,11 @@ sub _find_sig {
   my $sig = $_[0];
   return undef
     if !defined $sig;
-  local $@;
   return $sig
-    if ref $sig && eval { \&{$sig} };
+    if ref $sig;
   return undef
     if $sig eq 'DEFAULT' || $sig eq 'IGNORE';
+  # this isn't really needed because %SIG entries are always fully qualified
   package #hide
     main;
   no strict 'refs';
@@ -171,7 +219,7 @@ sub _warn {
   local $SIG{__WARN__};
   my @convert = _convert(@_);
   if (my $sig = _find_sig($OLD_SIG{__WARN__})) {
-    $sig->(ref $convert[0] ? $convert[0] : join('', @convert));
+    (\&$sig)->(ref $convert[0] ? $convert[0] : join('', @convert));
   }
   else {
     @convert = _ex_as_strings(@convert);
@@ -183,7 +231,7 @@ sub _die {
   local $SIG{__DIE__};
   my @convert = _convert(@_);
   if (my $sig = _find_sig($OLD_SIG{__DIE__})) {
-    $sig->(ref $convert[0] ? $convert[0] : join('', @convert));
+    (\&$sig)->(ref $convert[0] ? $convert[0] : join('', @convert));
   }
   @convert = _ex_as_strings(@convert) if _can_stringify;
   @convert = _colorize(31, @convert) if $OPTIONS{color} && _can_stringify;
@@ -198,7 +246,7 @@ sub _die {
 
 sub _colorize {
   my ($color, @convert) = @_;
-  if ($ENV{DEVEL_CONFESS_FORCE_COLOR} || -t *STDERR) {
+  if ($OPTIONS{color} eq 'force' || -t *STDERR) {
     if (@convert == 1) {
       $convert[0] = s/(.*)//;
       unshift @convert, $1;
@@ -227,12 +275,15 @@ sub _stack_trace {
     = $OPTIONS{dump} ? \&_ref_formatter : \&_str_val;
   my $message = &longmess;
   $message =~ s/\.?$/./m;
-  if ($OPTIONS{source}) {
-    $message .= Devel::Confess::Source::source_trace(1);
+  if ($OPTIONS{source} || $OPTIONS{evalsource}) {
+    $message .= Devel::Confess::Source::source_trace(1,
+      $OPTIONS{evalsource} ? ($OPTIONS{evalsource}, 1) : $OPTIONS{source});
   }
   $message;
 }
 
+# these are package varibles to control their lifetime.  they should not be
+# used externally.
 our $PACK_SUFFIX = 'A000';
 
 our %EXCEPTIONS;
@@ -332,29 +383,41 @@ sub _convert {
 
     return $ex;
   }
-  elsif ((caller(1))[0] eq 'Carp') {
-    my $out = join('', @_);
 
+  my $out = join('', @_);
+
+  if (caller(1) eq 'Carp') {
     my $long = longmess();
     my $long_trail = $long;
     $long_trail =~ s/.*?\n//;
     $out =~ s/\Q$long\E\z|\Q$long_trail\E\z//
       or $out =~ s/(.*) at .*? line .*?\n\z/$1/;
+  }
 
-    return ($out, _stack_trace());
+  my $source_trace;
+  $out =~ s/^(={75}\ncontext for .*^={75}\n\z)//ms
+    and $source_trace = $1;
+  my $trace = _stack_trace();
+  $trace =~ s/^(.*\n?)//;
+  my $where = $1;
+  my $new_source_trace;
+  $trace =~ s/^(={75}\ncontext for .*^={75}\n\z)//ms
+    and $new_source_trace = $1;
+  my $find = $where;
+  $find =~ s/(\.?\n?)\z//;
+  $out =~ s/(\Q$find\E(?: during global destruction)?(\.?\n?))(?:\Q$trace\E)?\z//
+    and $where = $1;
+  if (defined $source_trace) {
+    if (defined $new_source_trace) {
+      $new_source_trace =~ s/^={75}\n//;
+      $source_trace =~ s/^(([-=])\2{74}\n)(?:\Q$new_source_trace\E)?\z/$1/ms;
+    }
+    $trace .= $source_trace;
   }
-  else {
-    my $message = _stack_trace();
-    $message =~ s/^(.*\n?)//;
-    my $where = $1;
-    my $find = $where;
-    $find =~ s/(\.?\n?)\z//;
-    $find = qr/\Q$find\E(?: during global destruction)?(\.?\n?)/;
-    my $out = join('', @_);
-    $out =~ s/($find)\z//
-      and $where = $1;
-    return ($out, $where . $message);
+  if (defined $new_source_trace) {
+    $trace .= $new_source_trace;
   }
+  return ($out, $where . $trace);
 }
 
 sub _ex_as_strings {
@@ -485,13 +548,13 @@ Can also be used inside a script:
 =head1 DESCRIPTION
 
 This module is meant as a debugging aid. It can be used to make a script
-complain loudly with stack backtraces when warn()ing or die()ing.  Unlike other
-similar modules (e.g. L<Carp::Always>), it includes stack traces even when
-exception objects are thrown.
+complain loudly with stack backtraces when C<warn()>ing or C<die()>ing.
+Unlike other similar modules (e.g. L<Carp::Always>), stack traces will also be
+included when exception objects are thrown.
 
-The stack traces are generated using L<Carp>, and will look work for all types
-of errors.  L<Carp>'s C<carp> and C<confess> functions will also be made to
-include stack traces.
+The stack traces are generated using L<Carp>, and will work for all types of
+errors.  L<Carp>'s C<carp> and C<croak> functions will also be made to include
+stack traces.
 
   # it works for explicit die's and warn's
   $ perl -d:Confess -e 'sub f { die "arghh" }; sub g { f }; g'
@@ -506,11 +569,16 @@ include stack traces.
           main::f(undef) called at -e line 2
           main::g() called at -e line 2
 
-Internally, this is implemented with C<$SIG{__WARN__}> and C<$SIG{__DIE__}>
-hooks.
+Internally, this is implemented with L<$SIG{__WARN__}|perlvar/%SIG> and
+L<$SIG{__DIE__}|perlvar/%SIG> hooks.
 
 Stack traces are also included if raw non-object references are thrown.
 
+This module is compatible with all perl versions back to 5.6.2, without
+additional prerequisites.  It contains workarounds for a number of bugs in the
+perl interpreter, some of which effect comparatively simpler modules, like
+L<Carp::Always>.
+
 =head1 METHODS
 
 =head2 import( @options )
@@ -532,7 +600,9 @@ stack traces on supported exception types.  Disabled by default.
 =item C<dump>
 
 Dumps the contents of references in arguments in stack trace, instead
-of only showing their stringified version.  Shows up to three references deep.
+of only showing their stringified version.  Also causes exceptions that are
+non-object references and objects without string overloads to be dumped if
+being displayed.  Shows up to three references deep.
 Disabled by default.
 
 =item C<dump0>, C<dump1>, C<dump2>, etc
@@ -549,6 +619,22 @@ Colorizes error messages in red and warnings in yellow.  Disabled by default.
 Includes a snippet of the source for each level of the stack trace. Disabled
 by default.
 
+=item C<source0>, C<source1>, C<source2>, etc
+
+Enables source display, but with a specified number of lines of context to show.
+Context of 0 will show the entire source of the files.
+
+=item C<evalsource>
+
+Similar to the source option, but only shows includes source for string evals.
+Useful for seeing the results of code generation.  Disabled by default.
+Overrides the source option.
+
+=item C<evalsource0>, C<evalsource1>, C<evalsource2>, etc
+
+Enables eval source display, but with a specified number of lines of context to
+show.  Context of 0 will show the entire source of the evals.
+
 =item C<better_names>
 
 Use more informative names to string evals and anonymous subs in stack
@@ -625,10 +711,7 @@ L<Carp::Always::Dump>
 
 =back
 
-Please report bugs via CPAN RT
-http://rt.cpan.org/NoAuth/Bugs.html?Dist=Devel-Confess.
-
-=head1 BUGS
+=head1 CAVEATS
 
 This module uses several ugly tricks to do its work and surely has bugs.
 
@@ -636,30 +719,48 @@ This module uses several ugly tricks to do its work and surely has bugs.
 
 =item *
 
-This module does not play well with other modules which fusses
-around with C<warn>, C<die>, C<$SIG{'__WARN__'}>,
-C<$SIG{'__DIE__'}>.
+This module uses C<$SIG{__WARN__}> and C<$SIG{__DIE__}> to accomplish its goal,
+and thus may not play well with other modules that try to use these hooks.
+Significant effort has gone into making this work as well as possible, but
+global variables like these can never be fully encapsulated.
+
+=item *
+
+To provide stack traces on exception objects, this module re-blesses the
+exception objects into a generated class.  While it tries to have the smallest
+effect it can, some things cannot be worked around.  In particular,
+C<ref($exception)> will return a different value than may be expected.  Any
+module that relies on the specific return value from C<ref> like already has
+bugs though.
 
 =back
 
+=head1 SUPPORT
+
+Please report bugs via
+L<CPAN RT|http://rt.cpan.org/NoAuth/Bugs.html?Dist=Devel-Confess>.
+
+
 =head1 AUTHORS
 
-=over
+=over 4
 
 =item *
 
 Graham Knop <haarg@haarg.org>
 
+=back
+
+=head1 CONTRIBUTORS
+
+=over 4
+
 =item *
 
 Adriano Ferreira <ferreira@cpan.org>
 
 =back
 
-=head1 CONTRIBUTORS
-
-None yet.
-
 =head1 COPYRIGHT
 
 Copyright (c) 2005-2013 the L</AUTHORS> and L</CONTRIBUTORS>
